// This program demonstrates a practical use case of a "best effort" write-once memory
// The basic code block supports less guaranteed writes than will actually be filled.
// By building in an efficient detection for failed writes and by using the statistical
// properties of the data, we can achieve a high effective coding rate while keeping the
// probability of failed writes negligible
// The unaltered example shows how to achieve a fixed rate of 2.27273 with t=10, using a
// 16 bit code that is optimized to support a high average number of writes for 16 symbols.
// See report "A heuristic approach to find short efficient WOM codes"
// Comments to bert.o.dobbelaere[at]telenet[dot]be


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <time.h>
#include <math.h>
#include "codetable.h"


#define MAXNWRITES 20 

typedef unsigned char u8;
typedef unsigned u32;



u8 num_guaranteed_writes[NCOMBS];
double avgwrites[NCOMBS];


static Pattern_t decisiontable[NCOMBS][NSYMBOLS];

// Iterate over patterns while all bits from mask remain set.
u32 next_pattern(u32 p, u32 mask) 
{
	u32 i=1;
	p++;
	while((p&mask)!=mask)
	{
		if((~p&mask)&i)
			p+=i;
		i*=2;
	}	
	return p;
}


Pattern_t slowencode(Pattern_t pattern, Symbol_t symbol)
{
	Pattern_t bestpattern=DEADCODE; // not found
	int nwrites=-1;
	double avgexp=0;
	
	for(u32 p=pattern;p<NCOMBS;p=next_pattern(p,pattern))
		if(wom_decode[p]==symbol)
		{
			if((int)num_guaranteed_writes[p]>nwrites)
			{
				nwrites=num_guaranteed_writes[p];
				avgexp=avgwrites[p];
				bestpattern=p;
			}
			else if (((int)num_guaranteed_writes[p]==nwrites) && (avgwrites[p]>avgexp))
			{
				avgexp=avgwrites[p];
				bestpattern=p;
			}
		}
	
	return bestpattern;
}


void populate_decisions()
{
	printf("Populating decision table (trading memory for speed)...\n");
	for(u32 p=0;p<NCOMBS;p++)
		for(u32 s=0;s<NSYMBOLS;s++)
			decisiontable[p][s]=slowencode(p,s);
	
}


void initEncodingTable()
{
	printf("Analyzing decoding table...\n");
	for(int p=NCOMBS-2;p>=0;p--)
	{
		u8 highestfound[NSYMBOLS];
		double avgnext[NSYMBOLS];
		memset(highestfound,0,sizeof(highestfound));
		memset(avgnext,0,sizeof(avgnext));
		for(int q=next_pattern(p,p);q<(NCOMBS-1);q=next_pattern(q,p))
			{
				u8 s=wom_decode[q];
				if(highestfound[s]<num_guaranteed_writes[q])
					highestfound[s]=num_guaranteed_writes[q];
			}
		
		u8 nwrites=MAXNWRITES;
		for(int s=0;s<NSYMBOLS;s++)
			if((s!=wom_decode[p]) && (highestfound[s]<nwrites))
				nwrites=highestfound[s];
		num_guaranteed_writes[p] = nwrites+1;
		for(int q=next_pattern(p,p);q<(NCOMBS-1);q=next_pattern(q,p))
			{
				u8 s=wom_decode[q];
				if(avgnext[s]<(avgwrites[q]+1))
					avgnext[s]=(avgwrites[q]+1);
			}
		avgwrites[p]=1;
		for(int s=0;s<NSYMBOLS;s++)
			if(s!=wom_decode[p])
				avgwrites[p]+=avgnext[s];
		avgwrites[p]/=(NSYMBOLS-1);
	}
	
	for(int p=0;p<(NCOMBS-1);p++)
		num_guaranteed_writes[p]--; // counted 1 too many
	
	
	printf("Detected guaranteed writes = %u\n",num_guaranteed_writes[0]);   
	assert(num_guaranteed_writes[0]>=GUARANTEEDWRITES);
	populate_decisions();
}





u32 encode(u32 pattern, u8 symbol)
{
	return decisiontable[pattern][symbol];
}




void blockdemo()
{
	const u32 NTRIALS=10000000;
	
	printf("\n\nStart trials on individual code block. We write new random symbols to the block as long as they can be written.\n");
	printf("This process will be repeated %u times from an empty block...\n",NTRIALS);
	
	u32 writestats[MAXNWRITES]={0};
	
	
	for(u32 numtests=0;numtests<NTRIALS;numtests++)
	{
		u32 nwrites=0;
		u32 pattern=0;
		
		while(pattern<NCOMBS)
		{	
			u8 s=rand()%NSYMBOLS;	// Pick random symbol
			u32 pnext=encode(pattern,s);
			
			if(pnext==DEADCODE) 
				break; // Write failed 
			
			assert((pnext&pattern)==pattern); // What is written shall stay written
			
			assert( wom_decode[pnext] == s); // Pattern must decode to encoded symbol
			
			nwrites++;
			pattern=pnext;
		}
				
		if(nwrites<MAXNWRITES)
			writestats[nwrites]++;
		else
			writestats[MAXNWRITES-1]++;
		
		
		assert(nwrites>=GUARANTEEDWRITES);
		
	}
	
	double total=0.0;
	printf("Histogram - number of successful writes per trial:\n");
	printf(" N writes     Count\n");
	for(u32 n=0;n<(MAXNWRITES-1);n++)
	{
		printf("    %2u    %10u \n",n,writestats[n]);
		total+=n*writestats[n];
	}
	printf("   >%2u    %10u \n",MAXNWRITES-2,writestats[MAXNWRITES-1]);
	printf(" total    %10u\n",NTRIALS);
	printf("Average writes : %lf\n",total/NTRIALS);
	
	double cumulrate=0.0;
	double samples=NTRIALS;
	
	for(u32 n=1;n<=MAXNWRITES;n++)
	{
		samples-=writestats[n-1];
		double writerate = (samples/NTRIALS) * (log(NSYMBOLS)/log(2.0))/NBIT;
		cumulrate+=writerate;
		
		printf(" Write %2u :   Effective rate (this write):  %.6lf     Effective rate (cumul):  %.6lf\n",n,writerate,cumulrate);
	}
		
}

/**
 * Implements a sequential access WOM memory composed of a number of "best effort" code blocks
 */
class WOM_Memory
{
	public:
		WOM_Memory(size_t n_items);
		~WOM_Memory();
		void discardData();
		void resetReadPointer();
		bool readSymbol(Symbol_t &s);
		bool writeSymbol(Symbol_t s);
		void printStat() const;
		// No erase() implemented. Memory is considered a true WOM.
	private:
		Pattern_t *block_storage;
		size_t nblocks;
		size_t startindex;
		size_t write_offset;
		size_t read_offset;
};


WOM_Memory::WOM_Memory(size_t n_items)
{
	nblocks=n_items;
	block_storage = new Pattern_t[n_items];
	memset(block_storage, 0, n_items * sizeof(Pattern_t)); // Only place where '0' bits are written is during init.
	startindex=0;
	write_offset=0;
	read_offset=0;
}

WOM_Memory::~WOM_Memory()
{
	delete[] block_storage;
}

void WOM_Memory::discardData()
{
	startindex=(startindex+write_offset)%nblocks;
	write_offset=0;
	read_offset=0;
}

void WOM_Memory::resetReadPointer()
{
	read_offset=0;
}

bool WOM_Memory::readSymbol(Symbol_t &s)
{
	while(read_offset<write_offset)
	{
		size_t idx = (startindex+read_offset)%nblocks;
		Pattern_t p=block_storage[idx];
		read_offset++;
		
		if(p!=DEADCODE)
		{
			s=wom_decode[p];
			return true;
		}
	}
	return false;
}


bool WOM_Memory::writeSymbol(Symbol_t s)
{
	while(write_offset<nblocks)
	{
		size_t idx = (startindex+write_offset)%nblocks;
		Pattern_t p_read  = block_storage[idx];
		Pattern_t p_write = encode(p_read, s);
		assert((p_write & p_read) == p_read); // no cheating
		block_storage[idx] = p_write;
		
		write_offset++;
		if(p_write!=DEADCODE)
		{
			return true;
		}
	}
	
	printf("!! Memory write failure !!\n");
	return false;
}

void WOM_Memory::printStat() const
{
	size_t stopindex=(startindex+write_offset+nblocks-1)%nblocks;
	if(stopindex>=startindex)
		printf("Memory uses %lu of %lu blocks.  (locations %lu...%lu)\n", write_offset, nblocks, startindex, stopindex);
	else
		printf("Memory uses %lu of %lu blocks.  (locations %lu...%lu,  0...%lu)\n", write_offset, nblocks, startindex, nblocks-1, stopindex);	
}



void memorydemo()
{
	const u32 physical_blocks = 10800;
	const u32 data_blocks = 10000;
	const u32 nrounds = 10;
	
	printf("\nNow we create a WOM memory with a size of %u code blocks (%u bit each), ",physical_blocks, NBIT);
	printf("in which we will attempt to store\n%u times %u random symbols in range [0..%u]. ",nrounds,data_blocks,NSYMBOLS-1);
	printf("The WOM memory should have more physical blocks\nthan expected data symbols, to compensate for exhausted blocks near the end of the writes.\n");
	printf("The input data is assumed to have a high entropy, so the statistical distribution of exhausted blocks is well known.\n\n");
	
	
	WOM_Memory wom(physical_blocks);
	
	Symbol_t *data = new Symbol_t[data_blocks];
	bool rv;
	for(u32 round=1;round<=nrounds;round++)
	{
		printf("Round %u\n",round);
		
		// fill data
		for(u32 n=0;n<data_blocks;n++)
			data[n]=rand()%NSYMBOLS;
		
		// write data
		for(u32 n=0;n<data_blocks;n++)
		{
			rv=wom.writeSymbol(data[n]);
			if(rv==false)
			{
				printf("Write did not succeed. (n=%u)\n",n);
				break; 
			}
		}
		if(rv==false)
			break;
		
		// read and compare
		wom.resetReadPointer();
		
		for(u32 n=0;n<data_blocks;n++)
		{
			Symbol_t s;
			rv=wom.readSymbol(s);
			if(rv==false)
			{
				printf("Read did not succeed. (n=%u)\n",n);
				break; 
			}
			
			assert( s == data[n]);
			
		}
		if(rv==false)
			break;
		
		wom.printStat();
		
		// Data can now be overwritten
		wom.discardData();	
	}
	
	if(rv==true)
	{
		printf("\nEffective operating rate of memory as a whole: %lf (fixed) , t=%u\n\n", nrounds*data_blocks*log2(NSYMBOLS) / (physical_blocks * NBIT), nrounds);
	}
	
	delete[] data;
}


int main()
{
	srand(clock());
	initEncodingTable();

	blockdemo();
	memorydemo();
	
	
	return 0;
}


