// Demonstrates how a "best effort" code can be used to achieve a high rate in cyclical
// storage applications (here "security video"). By using a mix of different generations,
// the 16 symbols / 16 bit code with failed write detection (see best_effort_demo example)
// can be used in a very effective way, achieving a, effective coding rate of above 2.5.
// See report "A heuristic approach to find short efficient WOM codes"
// Comments to bert.o.dobbelaere[at]telenet[dot]be


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <time.h>
#include <math.h>
#include "codetable.h"


#define MAXNWRITES 20 

typedef unsigned char u8;
typedef unsigned u32;



u8 num_guaranteed_writes[NCOMBS];
double avgwrites[NCOMBS];


static Pattern_t decisiontable[NCOMBS][NSYMBOLS];

// Iterate over patterns while all bits from mask remain set.
u32 next_pattern(u32 p, u32 mask) 
{
	u32 i=1;
	p++;
	while((p&mask)!=mask)
	{
		if((~p&mask)&i)
			p+=i;
		i*=2;
	}	
	return p;
}


Pattern_t slowencode(Pattern_t pattern, Symbol_t symbol)
{
	Pattern_t bestpattern=DEADCODE; // not found
	int nwrites=-1;
	double avgexp=0;
	
	for(u32 p=pattern;p<NCOMBS;p=next_pattern(p,pattern))
		if(wom_decode[p]==symbol)
		{
			if((int)num_guaranteed_writes[p]>nwrites)
			{
				nwrites=num_guaranteed_writes[p];
				avgexp=avgwrites[p];
				bestpattern=p;
			}
			else if (((int)num_guaranteed_writes[p]==nwrites) && (avgwrites[p]>avgexp))
			{
				avgexp=avgwrites[p];
				bestpattern=p;
			}
		}
	
	return bestpattern;
}


void populate_decisions()
{
	printf("Populating decision table (trading memory for speed)...\n");
	for(u32 p=0;p<NCOMBS;p++)
		for(u32 s=0;s<NSYMBOLS;s++)
			decisiontable[p][s]=slowencode(p,s);
	
}


void initEncodingTable()
{
	printf("Analyzing decoding table...\n");
	for(int p=NCOMBS-2;p>=0;p--)
	{
		u8 highestfound[NSYMBOLS];
		double avgnext[NSYMBOLS];
		memset(highestfound,0,sizeof(highestfound));
		memset(avgnext,0,sizeof(avgnext));
		for(int q=next_pattern(p,p);q<(NCOMBS-1);q=next_pattern(q,p))
			{
				u8 s=wom_decode[q];
				if(highestfound[s]<num_guaranteed_writes[q])
					highestfound[s]=num_guaranteed_writes[q];
			}
		
		u8 nwrites=MAXNWRITES;
		for(int s=0;s<NSYMBOLS;s++)
			if((s!=wom_decode[p]) && (highestfound[s]<nwrites))
				nwrites=highestfound[s];
		num_guaranteed_writes[p] = nwrites+1;
		for(int q=next_pattern(p,p);q<(NCOMBS-1);q=next_pattern(q,p))
			{
				u8 s=wom_decode[q];
				if(avgnext[s]<(avgwrites[q]+1))
					avgnext[s]=(avgwrites[q]+1);
			}
		avgwrites[p]=1;
		for(int s=0;s<NSYMBOLS;s++)
			if(s!=wom_decode[p])
				avgwrites[p]+=avgnext[s];
		avgwrites[p]/=(NSYMBOLS-1);
	}
	
	for(int p=0;p<(NCOMBS-1);p++)
		num_guaranteed_writes[p]--; // counted 1 too many
	
	
	printf("Detected guaranteed writes = %u\n",num_guaranteed_writes[0]);   
	assert(num_guaranteed_writes[0]>=GUARANTEEDWRITES);
	populate_decisions();
}





u32 encode(u32 pattern, u8 symbol)
{
	return decisiontable[pattern][symbol];
}




/**
 * Models an SLC flash sector.
 * While individual code blocks (16 bits each) can be written (WOM), erasure is an operation that
 * affects the whole sector and reduces the lifetime of the device.
 */
class FlashSector
{
	public:
		FlashSector(u32 id_number, size_t n_blocks);
		~FlashSector();
		bool readSymbol(Symbol_t &s);
		bool writeSymbol(Symbol_t s);
		void discardData();
		size_t getSymbolsStored() const;
		void erase();
	private:
		Pattern_t *block_storage;
		u32 id;
		size_t nblocks;
		size_t write_offset;
		size_t read_offset;
		size_t symbols_stored;
};


FlashSector::FlashSector(u32 id_number, size_t n_blocks)
{
	nblocks=n_blocks;
	block_storage = new Pattern_t[n_blocks];
	memset(block_storage, 0, n_blocks * sizeof(Pattern_t));
	write_offset=0;
	read_offset=0;
	symbols_stored=0;
	id=id_number;
}

FlashSector::~FlashSector()
{
	delete[] block_storage;
}

size_t FlashSector::getSymbolsStored() const
{
	return symbols_stored;
}

void FlashSector::discardData()
{
	symbols_stored=0;
	write_offset=0;
	read_offset=0;
}

bool FlashSector::readSymbol(Symbol_t &s)
{
	while(read_offset<nblocks)
	{
		size_t idx = read_offset;
		Pattern_t p=block_storage[idx];
		read_offset++;
		
		if(p!=DEADCODE)
		{
			s=wom_decode[p];
			return true;
		}
	}
	return false;
}


bool FlashSector::writeSymbol(Symbol_t s)
{
	while(write_offset<nblocks)
	{
		size_t idx = write_offset;
		Pattern_t p_read  = block_storage[idx];
		Pattern_t p_write = encode(p_read, s);
		assert((p_write & p_read) == p_read); // no cheating
		block_storage[idx] = p_write;
		
		write_offset++;
		if(p_write!=DEADCODE)
		{
			symbols_stored++;
			return true;
		}
	}
	
	return false;
}

void FlashSector::erase()
{
	printf("Erasing flash sector %u\n",id);
	memset(block_storage, 0, nblocks * sizeof(Pattern_t));
	discardData();
}



void video_demo()
{
	const u32 blocks_per_sector  = 1000;
	const u32 num_write_attempts = 11;
	const u32 sectors_per_level  = 10;
	const u32 data_symbols       = 2000000; // Need to write sufficient data to exit the initial regime in which all writes succeed.
	
	const u32 totalsectors = num_write_attempts * sectors_per_level;
	
	printf("In this example we assume a surveillance video application, using an SLC flash memory.\n");
	printf("The video system uses a cyclical buffer, which overwrites the oldest data. Erasing flash memory\n");
	printf("often reduces its lifetime. The intention is to reduce the number of sector erase operations.\n");
	printf("Each flash sector is erased in one atomic operation and has a size of %u %u-bit code blocks.\n",blocks_per_sector,NBIT);
	printf("The memory contains %u sectors, and each pass a fraction 1/%uth of the sectors will be erased.\n",totalsectors,num_write_attempts);
	printf("When writing a sector, it is filled with the maximum amount of symbols in range [0..%u] it can store.\n",NSYMBOLS-1);
	printf("This amount will be %u for the first %u writes, and is reduced for the higher write numbers.\n",blocks_per_sector, GUARANTEEDWRITES);
	printf("Due to the fact that the written data will be distributed equally over the write numbers [1..%u],\n",num_write_attempts);
	printf("the amount of data stored will be fairly constant, although subject to statistics, but in any case decodable.\n");
	printf("The example assumes the stored data has a high entropy, which is normally the case when storing compressed video.\n\n");  
	
	printf("Creating flash model...\n");	
	FlashSector **sector = new FlashSector* [totalsectors];
		
	for(u32 id=0;id<totalsectors;id++)
	{
		sector[id] = new FlashSector(id, blocks_per_sector);
	}
	
	
	// Fill video data with random symbols. This is the input to be recorded in the cyclical buffer
	printf("Initializing sample data.\n");
	
	Symbol_t *data = new Symbol_t[data_symbols];
	for(u32 n=0;n<data_symbols;n++)
		data[n]=rand()%NSYMBOLS;
	
	// Now write all the data to the cyclical system
	
	u32 data_idx=0;
	u32 current_sector=0;
	u32 sector_loopcount=0;
	u32 oldest_data_index=0;
	
	printf("Running 'recording' cycle, writing all data and performing periodic sector erase operations.\n");
	while(data_idx<data_symbols)
	{
		bool write_ok;
		do
		{
			write_ok = sector[current_sector]->writeSymbol( data[data_idx] );
			if(write_ok)
				data_idx++;
			
		} while(write_ok && (data_idx<data_symbols));
#if 0		
		printf("Wrote %lu symbols in sector %u\n", sector[current_sector]->getSymbolsStored(), current_sector);
#endif		
		current_sector++;
		if(current_sector>=totalsectors)
		{
			current_sector=0;
			sector_loopcount++;
		}
			
		if(data_idx<data_symbols)
		{
			oldest_data_index += sector[current_sector]->getSymbolsStored();
			sector[current_sector]->discardData();
			
			// Erase sectors in an interleaved way, leaving num_write_attempts writes between successive erase cycles of the same sector.
			if( (current_sector+sector_loopcount) % num_write_attempts == 0)
			{
				sector[current_sector]->erase();
			}
			
		}
		
	}
	
	// Read back the data in the buffer and compare with written contents.
	
	printf("\nIndex of oldest data in the buffer: %u. Reading back data and verifying.\n",oldest_data_index);
	data_idx=oldest_data_index;
	while(data_idx<data_symbols)
	{
		bool read_ok;
		do
		{
			Symbol_t s;
			read_ok = sector[current_sector]->readSymbol( s );
			if(read_ok)
			{
				// Verify that we read back the written symbol
				assert( s == data[data_idx] );
				
				data_idx++;
			}
		}while(read_ok && (data_idx<data_symbols));

		if(data_idx<data_symbols)
		{
			current_sector++;
			if(current_sector>=totalsectors)
			{
				current_sector=0;
			}			
		}
	}
	
	u32 nsymbols_buffered = data_symbols-oldest_data_index;
	double stored_rate = log2(NSYMBOLS)*nsymbols_buffered / (NBIT * totalsectors * blocks_per_sector);
	printf("The buffer memory holds the last %u data symbols, stored in %u code blocks of %u bit.\n", nsymbols_buffered, totalsectors * blocks_per_sector,NBIT);
	printf("The effective rate of the last write is %lf. Sectors are only erased each %u writes,\n", stored_rate, num_write_attempts);
	printf("which corresponds to a total rate of %lf\n", stored_rate * num_write_attempts);
	
	
	delete[] data;

	for(u32 id=0;id<totalsectors;id++)
		delete sector[id];
	delete[] sector;

}


int main()
{
	srand(clock());
	initEncodingTable();

	video_demo();
	
	
	return 0;
}


